=== modified file 'bzrlib/commit.py'
--- bzrlib/commit.py	2006-06-07 13:39:33 +0000
+++ bzrlib/commit.py	2006-06-07 14:56:37 +0000
@@ -324,10 +324,10 @@
                         self.specific_files)
             self._check_parents_present()
             self.builder = self.branch.get_commit_builder(self.parents)
+            self.builder.set_revision_id(self.rev_id)
             
             self._remove_deleted()
             self._populate_new_inv()
-            self._store_snapshot()
             self._report_deletes()
 
             if not (self.allow_pointless
@@ -336,11 +336,7 @@
                 raise PointlessCommit()
 
             self._emit_progress_update()
-            self.inv_sha1 = self.branch.repository.add_inventory(
-                self.rev_id,
-                self.builder.new_inventory,
-                self.present_parents
-                )
+            self.inv_sha1 = self.builder.finish_inventory()
             self._emit_progress_update()
             self._make_revision()
             # revision data is in the local branch now.
@@ -480,10 +476,11 @@
 
     def _gather_parents(self):
         """Record the parents of a merge for merge detection."""
+        # TODO: Make sure that this list doesn't contain duplicate 
+        # entries and the order is preserved when doing this.
         pending_merges = self.work_tree.pending_merges()
         self.parents = []
         self.parent_invs = []
-        self.present_parents = []
         precursor_id = self.branch.last_revision()
         if precursor_id:
             self.parents.append(precursor_id)
@@ -492,7 +489,6 @@
             if self.branch.repository.has_revision(revision):
                 inventory = self.branch.repository.get_inventory(revision)
                 self.parent_invs.append(inventory)
-                self.present_parents.append(revision)
 
     def _check_parents_present(self):
         for parent_id in self.parents:
@@ -542,38 +538,6 @@
                 del self.work_inv[file_id]
             self.work_tree._write_inventory(self.work_inv)
 
-    def _store_snapshot(self):
-        """Pass over inventory and record a snapshot.
-
-        Entries get a new revision when they are modified in 
-        any way, which includes a merge with a new set of
-        parents that have the same entry. 
-        """
-        # XXX: Need to think more here about when the user has
-        # made a specific decision on a particular value -- c.f.
-        # mark-merge.  
-
-        # iter_entries does not visit the ROOT_ID node so we need to call
-        # self._emit_progress_update once by hand.
-        self._emit_progress_update()
-        for path, ie in self.builder.new_inventory.iter_entries():
-            self._emit_progress_update()
-            self.builder.record_entry_contents(ie, self.parent_invs, 
-                self.rev_id, path, self.work_tree)
-            # describe the nature of the change that has occured relative to
-            # the basis inventory.
-            if (self.basis_inv.has_id(ie.file_id)):
-                basis_ie = self.basis_inv[ie.file_id]
-            else:
-                basis_ie = None
-            change = ie.describe_change(basis_ie, ie)
-            if change in (InventoryEntry.RENAMED, 
-                InventoryEntry.MODIFIED_AND_RENAMED):
-                old_path = self.basis_inv.id2path(ie.file_id)
-                self.reporter.renamed(change, old_path, path)
-            else:
-                self.reporter.snapshot_change(change, path)
-
     def _populate_new_inv(self):
         """Build revision inventory.
 
@@ -589,38 +553,46 @@
         # iter_entries does not visit the ROOT_ID node so we need to call
         # self._emit_progress_update once by hand.
         self._emit_progress_update()
+        self._emit_progress_update()
         for path, new_ie in self.work_inv.iter_entries():
             self._emit_progress_update()
+            self._emit_progress_update()
             file_id = new_ie.file_id
             mutter('check %s {%s}', path, new_ie.file_id)
             if (not self.specific_files or 
                 is_inside_or_parent_of_any(self.specific_files, path)):
                     mutter('%s selected for commit', path)
-                    self._select_entry(new_ie)
+                    ie = new_ie.copy()
+                    ie.revision = None
             else:
                 mutter('%s not selected for commit', path)
-                self._carry_entry(file_id)
+                if self.basis_inv.has_id(file_id):
+                    ie = self.basis_inv[file_id].copy()
+                else:
+                    # this entry is new and not being committed
+                    continue
+
+            self.builder.record_entry_contents(ie, self.parent_invs, 
+                self.rev_id, path, self.work_tree)
+            # describe the nature of the change that has occured relative to
+            # the basis inventory.
+            if (self.basis_inv.has_id(ie.file_id)):
+                basis_ie = self.basis_inv[ie.file_id]
+            else:
+                basis_ie = None
+            change = ie.describe_change(basis_ie, ie)
+            if change in (InventoryEntry.RENAMED, 
+                InventoryEntry.MODIFIED_AND_RENAMED):
+                old_path = self.basis_inv.id2path(ie.file_id)
+                self.reporter.renamed(change, old_path, path)
+            else:
+                self.reporter.snapshot_change(change, path)
 
     def _emit_progress_update(self):
         """Emit an update to the progress bar."""
         self.pb.update("Committing", self.pb_count, self.pb_total)
         self.pb_count += 1
 
-    def _select_entry(self, new_ie):
-        """Make new_ie be considered for committing."""
-        ie = new_ie.copy()
-        ie.revision = None
-        self.builder.new_inventory.add(ie)
-        return ie
-
-    def _carry_entry(self, file_id):
-        """Carry the file unchanged from the basis revision."""
-        if self.basis_inv.has_id(file_id):
-            self.builder.new_inventory.add(self.basis_inv[file_id].copy())
-        else:
-            # this entry is new and not being committed
-            self.pb_total -= 1
-
     def _report_deletes(self):
         for path, ie in self.basis_inv.iter_entries():
             if ie.file_id not in self.builder.new_inventory:

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2006-06-07 12:19:04 +0000
+++ bzrlib/repository.py	2006-06-07 14:56:37 +0000
@@ -69,9 +69,17 @@
         inv_sha1 = bzrlib.osutils.sha_string(inv_text)
         inv_vf = self.control_weaves.get_weave('inventory',
                                                self.get_transaction())
-        inv_vf.add_lines(revid, parents, bzrlib.osutils.split_lines(inv_text))
+        self._inventory_add_lines(inv_vf, revid, parents, bzrlib.osutils.split_lines(inv_text))
         return inv_sha1
 
+    def _inventory_add_lines(self, inv_vf, revid, parents, lines):
+        final_parents = []
+        for parent in parents:
+            if parent in inv_vf:
+                final_parents.append(parent)
+
+        inv_vf.add_lines(revid, final_parents, lines)
+
     @needs_write_lock
     def add_revision(self, rev_id, rev, inv=None, config=None):
         """Add rev to the revision store as rev_id.
@@ -222,7 +230,7 @@
 
     def get_commit_builder(self, branch, parents):
         """Obtain a CommitBuilder for this repository."""
-        return CommitBuilder(self)
+        return CommitBuilder(self, parents)
 
     def unlock(self):
         self.control_files.unlock()
@@ -705,6 +713,9 @@
 class KnitRepository(MetaDirRepository):
     """Knit format repository."""
 
+    def _inventory_add_lines(self, inv_vf, revid, parents, lines):
+        inv_vf.add_lines_with_ghosts(revid, parents, lines)
+
     @needs_read_lock
     def all_revision_ids(self):
         """See Repository.all_revision_ids()."""
@@ -1848,9 +1859,21 @@
     know the internals of the format of the repository.
     """
 
-    def __init__(self, repository):
+    def __init__(self, repository, parents):
+        self.parents = parents
+        self.new_inventory = Inventory()
         self.repository = repository
-        self.new_inventory = Inventory()
+
+    def finish_inventory(self):
+        """Tell the builder that the inventory is finished.
+        
+        :return: SHA1 of the encoded inventory.
+        """
+        return self.repository.add_inventory(
+            self._new_revision_id,
+            self.new_inventory,
+            self.parents
+            )
 
     def set_revision_id(self, revision_id):
         """Set the revision id for this commit.
@@ -1882,17 +1905,21 @@
         # be made to force a revision, which will fail when they cannot be set
         # in this manner.
         self._new_revision_id = revision_id
+
+        self.new_inventory.add(ie)
+
         # ie.revision is always None if the InventoryEntry is considered
         # for committing. ie.snapshot will record the correct revision 
         # which may be the sole parent if it is untouched.
+        if ie.revision is not None:
+            return
         previous_entries = ie.find_previous_heads(
             parent_invs,
             self.repository.weave_store,
             self.repository.get_transaction())
-        if ie.revision is None:
-            # we are creating a new revision for ie in the history store
-            # and inventory.
-            ie.snapshot(revision_id, path, previous_entries, tree, self)
+        # we are creating a new revision for ie in the history store
+        # and inventory.
+        ie.snapshot(revision_id, path, previous_entries, tree, self)
 
     def modified_directory(self, file_id, file_parents):
         """Record the presence of a symbolic link.

